#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <sys/ipc.h> 
#include <sys/shm.h> 
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
 
#include <tcpip_lib.h>
 
#define     SERVER_PORT         6905
 
#define     SHD_MEM_KEY         6905
 
/* child stat */
#define     CHILD_STAT_IDL      0
#define     CHILD_STAT_BUSY     1
 
/* child shutdown */
#define     CHILD_NORMAL        0
#define     CHILD_SHUTDOWN      1
#define     CHILD_RESTART       2

/* parent control */
#define     PARENT_NORMAL           0
#define     PARENT_WAIT_DOWN        1
#define     PARENT_FORCE_DOWN       2
#define     PARENT_CHILD_RESTART    3

/*
* child  մϴ.
* child ȣǼ , 
*  ð  迡 ʿ  ߰  ֽϴ.
*/
typedef struct {
    int     pid;
    int     stat;
    int     shutdown;
    unsigned long call_count;
} child_info_t;

typedef struct {
    int     status;    /* θ μ  */
    child_info_t children[];
} process_info_t;

/*
* server port, shared memory key  ȯ漳Ͽ о 
* ϴ  
*/
int g_server_port   = SERVER_PORT;
int g_shd_mem_key   = SHD_MEM_KEY;

/*
* child process min, max   ƴ  Ͽ
*  ȯ Ͽ оų main parameter .
*/
int g_max_child_count = 200; /*ִ  child μ */
int g_min_child_count = 50;  /*ʿ  child μ */
int g_cur_child_count = 0;   /*  child μ */

int g_binded_socket;    /*   scriptor*/
int g_accepted_socket;  /* ӵ  scriptor*/

process_info_t  *g_process_info;
child_info_t    *g_children; /* ü child  */
child_info_t    *g_child;    /* child ڽ  */

void parent_main(void);
int  create_child(void);
void create_children(void);
void shutdown_all_children(int flag);
void children_control(void);
void child_main(void);
void child_job(void);
 
int main(int argc, char *argv[])
{
    int shmid;
    child_info_t    *child;
    int     idx;
    int     fd;
    int     mem_size;
    struct  sigaction   act_ignore;
	
    /* 
    * background program  terminal  signal ߻ . (Control-C ) 
    * ʽ α׷ Ͽ θ process id 1 child μ .
    *  session ׷ leader ƴ μ 
    */
    if(fork() != 0) {
        /* child process ƴϸ  */
        exit(0);
    }
    
    /*
    * ο session Ͽ session׷ leader 
    * ο session terminal   .
    */
    if(setsid() == -1) {
        fprintf(stderr, "NEW SESSION ERROR: %s\n", strerror(errno));
        exit(1);
    }
    
    /*
    * μ home directory ROOT丮 .
    */
    if(chdir("/") == -1) {
        fprintf(stderr, "CHDIR ERROR: %s\n", strerror(errno));
        exit(1);
    }

    /*
    * ǥԷ/ǥ/ǥؿ ó    ϱ Ͽ /dev/null  openմϴ.
    */
    if((fd = open("/dev/null", O_RDWR)) == -1) {
        exit(1);
    }
    
    /*
     * ǥԷ  ϵ redirect ó.   
     * scanf(), fgets(..., stdin)  Լ ϳ Է ȵ.
     */
    dup2(fd, 0); 
 
    /*
     * ǥ  ϵ redirect ó. 
     * printf()  Լ ϳ  ȵ
     */
    dup2(fd, 1); 
    
    /*
     * ǥؿ  ϵ redirect ó. 
     * fprintf(stderr, ...) Լ ϳ  ȵ
     */
    dup2(fd, 2); 
 
    close(fd);   /* ʿ fd close */
 
    memset(&act_ignore, 0x00, sizeof(act_ignore));
    act_ignore.sa_handler    = SIG_IGN;
    sigaction(SIGBUS,  &act_ignore, NULL);
    sigaction(SIGSEGV, &act_ignore, NULL);
    sigaction(SIGILL,  &act_ignore, NULL);
    sigaction(SIGFPE,  &act_ignore, NULL);


    /*
     * ̹ α׷ ̸ shared memory  ϸ ȵǱ 
     *  socket  .
     */
    if((g_binded_socket = TCPIPserver(g_server_port)) == -1) {
        /* TO-DO :  α  */
        exit(1);
    }

    /* accept socket μ µ üũؾϱ  NON-BLOCKING  */
    fcntl(g_binded_socket, F_SETFL, O_NONBLOCK);
    
    /* shared memory  */
    mem_size = sizeof(process_info_t) + sizeof(child_info_t) * g_max_child_count;
    shmid = shmget((key_t)g_shd_mem_key, mem_size, 0600 | IPC_CREAT);
    if (shmid == -1) {
        perror("shmget failed : ");
        exit(0);
    }
 
    /*  shared memroy  */
    g_process_info = shmat(shmid, (void *)0, 0);
    if (g_process_info == (void *)-1) {
        perror("shmat failed : ");
        exit(0);
    }
    
    memset(g_process_info, 0x00, mem_size);
    g_children = g_process_info->children;

    
    create_children();
 
    parent_main();
}

/*
 * 1. child process main θ μ  û  쿡 
 *     óĿ  ó.
 * 2. child process main client  ϰ child_job( )Լ ȣϿ
 *    clientû óϵ .
 */
void child_main(void)
{
    int acceted_socket;
    struct  sockaddr_in client;
 
    /* 
    θ μ CHILD_SHUTDOWN ϰų CHILD_RESTART ϸ μ .
    CHILD_SHUDOWN  ü μ ϰų ӽ÷  μ ϱ .
    CHILD_RESTART   ̺귯 ݿϰų Ÿ  ⵿ ʿ 
    */
    while(g_child->shutdown != CHILD_SHUTDOWN && g_child->shutdown != CHILD_RESTART) {
        g_child->stat     = CHILD_STAT_IDL;
        /* θμ Ǿ . */
        if(getppid() == 1) {
            exit(0);
        }
 
        if((g_accepted_socket = TCPIPaccept(g_binded_socket, &client)) == -1) {
            /* TO-DO :  α  */
            continue;
        }
 
        /* 
        * Ȥö g_accept_socket  child_job ٸ  Ͽ  
        * socket close ʴ   Ͽ accept_socket Ͽ close
        */
        acceted_socket = g_accepted_socket;
        g_child->stat     = CHILD_STAT_BUSY;
        g_child->call_count++;

        child_job();
 
        close(acceted_socket);
    }
 
    return;     
}

/*
 *  client  ͸ ְ ޴  ؾ մϴ.
 */
void child_job(void)
{
    /* TO-DO:  Լ    մϴ. */

#if 0    
    /* 
    * TO-DO:   °  ۼ ü Ͽ մϴ. 
    * ʿ  Ͽ մϴ.
    */
    int recv_len;
    char recv_buf[4096];
    char send_buf[4096];
    
    if((recv_len = TCPIPrecv(g_accepted_socket, recv_buf, 2048, 0)) == -1) {
        /* TO-DO :  α  */
        return ;
    }
    
    /* TO-DO : Է ͷ óϴ   */
    /* ...... */
    
    if((TCPIPsend(g_accepted_socket, send_buf, 2048, 0)) == -1) {
        /* TO-DO :  α  */
        return ;
    }
#endif
}

/*
 * ο child process մϴ.
 * ο child process  g_child ǹǷ 
 * create_child Լ ϱ  g_child child ġ ؾ .
 */
int create_child(void)
{
    pid_t   child_pid;
    
    if((child_pid = fork()) == -1) {
        return -1;
    } else if(child_pid == 0) {
        g_child->stat     = CHILD_STAT_IDL;
        g_child->shutdown = CHILD_NORMAL;
        child_main();
        exit(0);
    } else if(child_pid > 0) {
        g_child->pid = child_pid;
        return 0;
    }   
}

/*
 *   child process  
 * ο μ մϴ.
 */
void killed_children_check(void)
{
    int     idx;
    pid_t   child_pid;
    int     status;
    
    while(1) {
        if((child_pid = waitpid(-1, &status, WNOHANG)) == 0) {
            return;
        }
        if(child_pid == -1) {
            return;
        }
        for(idx = 0; idx < g_cur_child_count; idx++) {
            if(g_children[idx].pid != child_pid) {
                continue;
            }
            /*  Ǿ Ǿ  μ̸ */
            if(g_children[idx].shutdown == CHILD_NORMAL || g_children[idx].shutdown == CHILD_RESTART) {
                g_child = &g_children[idx];
                create_child();
                break;
            }
        }
    }   
}

/*
 * child μ  ۾̸, μ  ʿ  ǴϿ
 * ο μ մϴ.
 */
void busy_children_check(void)
{
    int idx;
    int running_child;
    child_info_t    *child;
    
    /* θ  child process ¸ ϸ鼭 process  ø ˻.*/
    running_child = 0;
 
    for(idx = 0; idx < g_cur_child_count; idx++) {
        child = &g_children[idx];
        /* process ̸ */
        
        if(child->stat == CHILD_STAT_BUSY) {
            running_child++;
        }
        
    }
 
    /*  ̸ */
    if(running_child == g_cur_child_count) {
        /*  max child count  ʾ, */
        if(g_cur_child_count < g_max_child_count) {
            g_child = &g_children[g_cur_child_count];
            /* ο child  */
            if(create_child() == 0) {
                g_cur_child_count++;
            }
        }
    }
}

/*
* θμ 1ʸ 
* 1. μ ϴ α׷ μ ٿ  ȴ  üũϿ
*    child reboot, , ó     .
* 2. child process  Ǿ  
*    ȮϿ   μ ٽ .
* 3. ִ μ  쿡 μ  ڶǷ
*    ο μ Ŵ(g_max_child_count   ̸)
*/
void parent_main(void)
{
    /* θ  child process  ϸ鼭 process  ø Ȯ.*/
    while(1) {
        sleep(1);
        
        children_control();
        killed_children_check();
        busy_children_check();
    }
}
 
/*  min ŭ child process  */
void create_children(void)
{
    int idx;
 
    for(g_cur_child_count = 0; g_cur_child_count < g_min_child_count; g_cur_child_count++) {
        g_child = &g_children[g_cur_child_count];
        if(create_child() == -1) {
            break;
        }
    }

    /* μ    ʾ, ü  */
    if(g_cur_child_count < g_min_child_count) {
        g_process_info->status = PARENT_FORCE_DOWN;
        children_control();
        exit(0);
    }
}

/*
* child process ü   restart  
* daemon  α׷(ۼ ʿ) shared memory ¸ Ͽ 
* ׿  child reboot, , ó     .
*/
void children_control(void)
{
    int idx;
    child_info_t    *child;
 
    if(g_process_info->status == PARENT_NORMAL) {
        return;
    }
 
    for(idx = 0; idx < g_cur_child_count; idx++) {
        child = &g_children[idx];
        
        if(g_process_info->status == PARENT_CHILD_RESTART) {
            /* child process ۾ ϷǸ, Ű ٽ Ŵ */
            child->shutdown = CHILD_RESTART;
        } else if(g_process_info->status == PARENT_FORCE_DOWN) {
            /* child process ۾ Ϸϱ⸦ ٸ ʰ   */
            child->shutdown = CHILD_SHUTDOWN;
            kill(child->pid, 9);
        } else if(g_process_info->status == PARENT_WAIT_DOWN) {
            /* child process ۾ Ǹ, Ŵ */
            child->shutdown = CHILD_SHUTDOWN;
        } else {
            g_process_info->status = PARENT_NORMAL;
            return;
        }
    }
    if(g_process_info->status != PARENT_CHILD_RESTART) {
        exit(0);
    }
    g_process_info->status = PARENT_NORMAL;
}
